from string import join
from cStringIO import StringIO

from OFS import SimpleItem
from Globals import DTMLFile
from Relations import Relations
from urllib import unquote

class Struct:
    __allow_access_to_unprotected_subobjects__=1
    def __init__(self, **entries): self.__dict__.update(entries)


#class MoveWrapper:
#
#
#    "Wrapping objects in this before they are moved, allows them to keep relations"
#
#
#    def __init__(self, ob, relationships):
#        # relationsips are a list of relationsship titles
#        self.relationships = relationships
#        self.__ob = ob
#
#
#    def __getattr__(self, name):
#        return getattr(self.__ob, name)
#
#
#    def _relationship_name(self, name)
#        receipt_prefix = '_receipts_for_'
#        return '%s%s' % (receipt_prefix, name)
#
#
#    def manage_beforeDelete(self, item, container):            
#        "Hook"
#        for relationship in self.relationships:
#            relation = self.relations[relationship]
#            relationship_name = self._relationship_name(relationship)
#            setattr(self, relationship_name) = relation.deposit(self)
#        self.__ob.manage_beforeDelete(self, item, container)
#    
#    
#    def manage_afterAdd(self, item, container):
#        "hook"
#        for relationship in self.relationships:
#            relation = self.relations[relationship]
#            relationship_name = self._relationship_name(relationship)
#            if hasattr(self, relationship_name):
#                relation.withdraw(self, getattr(self, relationship_name))
#                del(getattr(self, relationship_name))
#        self.__ob.manage_afterAdd(self, item, container)



class mxmRelations(SimpleItem.SimpleItem, Relations):

    """
    The mxmRelations product
    """

    meta_type = 'mxmRelations'

    manage_options = (
    	{'label': 'Properties', 'action': 'manage_editForm',},
    	{'label': 'View', 'action': 'index_html',},
    )

    def __init__(self, id, title):
        "Inits the product with default values"
        mxmRelations.inheritedAttribute('__init__')(self)
        self.id     = id
        self.title  = title

    def manage_editAction(self, title, RESPONSE=None):
        "Changes the product values"
        self.title      = title
        RESPONSE.redirect('manage_editForm')


    ##########################
    # The web pages that shows content. Put your own in the www folder.

    # Used to view content of the object
    index_html = DTMLFile('www/index_html', globals())

    # Edit the content of the object
    manage_editForm = DTMLFile('www/manage_editForm', globals())


    ########################################################
    # Helper methods


    def _o2path(self, object):
        "Returns a paths as a string"
#        return join(object.getPhysicalPath()[1:], '/')
        return join(object.getPhysicalPath()[1:], '/')


    def _getInput(self, objs1, objs2):
        """
        I couldn't think of a better name :-(, but it takes two lists
        of objects as input and returns two lists of paths to the objects
        """
        objs1 = self._toList(objs1)
        objs2 = self._toList(objs2)
        # convert objects to paths
        paths1 = self._objectsToPaths(objs1)
        paths2 = self._objectsToPaths(objs2)
        return (paths1, paths2)


    def _objectsToPaths(self, objList):
        """
        Takes a list of objects and returns a list of paths as strings
        of the form ['path/to/object/one','path/to/object/two']
        """
        return [self._o2path(obj) for obj in objList]


    def pathsToObjects(self, paths, meta_types=None):
        """
        returns a list of objects from a list of paths
        deletes any paths that are no longer valid because they have been
        deleted, moved, or renamed from Zope.
        """
        if meta_types != None:
            meta_types = self._toList(meta_types)
        result = []
        for path in paths:
            try:
                # get the object from the path
                path = unquote(path)
                obj = self.getPhysicalRoot().restrictedTraverse(path)
                # Filter out meta_type if nessecary
                if ((meta_types == None) or (obj.meta_type in meta_types)):
                    result.append(obj)
                else:
                    self._delete(path)
            except KeyError:
                # The object the path points to is deleted, moved,
                # or renamed from Zope.
                # So the relation is no longe valid. Delete it!
                self._delete(path)
            except 'Unauthorized':
                # If the user don't have access don't add the related object,
                # but don't delete the relation either
                pass
        return result

    ########################################################
    # Public methods.


    def relate(self, objs1, objs2):
        """
        Sets relations between objects.
        If there allready is a key it appende the relations
        else it creates a new key with a list of relations
        """
        paths1, paths2 = self._getInput(objs1, objs2)
        self._relate(paths1, paths2)


    def unrelate(self, objs1, objs2):
        """
        Removes relations between objects
        """
        paths1, paths2 = self._getInput(objs1, objs2)
        self._unrelate(paths1, paths2)


    def delete(self, obj):
        """
        Removes all references to the object.
        Used ie. if an object is deleted.
        """
        path = self._o2path(obj)
        self._delete(path)


    def get(self, obj, meta_types=None):
        """
        Returns all relations to this object, or an empty list
        """
        objectPath = self._o2path(obj)
        paths = self._get(objectPath)
        return self.pathsToObjects(paths, meta_types)


    def update(self, objs1, newObjs2):
        """
        Updates relations between objects. It deletes the old 
        relations not in newObjs2, and adds the new objects in 
        newObjs2. 
        """
        # it still needs to be written!
        pass


    def getAsPaths(self, obj, meta_types=None):
        """
        Returns all relations to this object as paths, or an empty list
        """
        objects = self.get(obj, meta_types)
        return self._objectsToPaths(objects)


    def relateByPaths(self, obj, pathsList):
        "Relates an object to a list of paths"
        self.smartRelate(obj, pathsList)


    #########################
    # used when relating from web form

    def getSmartList(self, obj, objects, meta_types=None):
        """
        returns a list of objects with attributes (id, title, path, selected)
        Very usable for making selection widgets.
        "obj" is the object that we are looking for relations too.
        "objects" is a list of objects that we want to know whether
        they are related to "obj".
        """
        result = []
        relations = self.get(obj, meta_types)
        for o in objects:
            if o in relations:
                selected = 1
            else:
                selected = 0
            result.append(
                Struct(
                    selected = selected,
                    path     = self._o2path(o),
                    title    = o.title_or_id(),
                    id       = o.getId()
                    )
                )
        return result


#    def smartRelate(self, obj, pathsList):
    def smartRelate(self, obj, pathsList=None):
        """
        Works in tandem with 'getSmartList' and takes what is generated
        by that and creates relations from them.
        """
#        # list with empty string can be returned by
#        # checkboxes with hidden '' as default
#        if pathsList == ['']:
#            pathsList = []
        if pathsList == None:
            pathsList = []
        self.delete(obj)
        objList = self.pathsToObjects(pathsList)
        self.relate(obj, objList)


    #########################
    # used when copy/pasting or moving objects

    def deposit(self, obj):
        """
        Used to return a receipt in a "manage_beforeDelete()" hook.
        Puts the relations of an object in a deposit, and returns a receipt.
        receipt = self.relations.deposit(self)
        """
        paths = self.getAsPaths(obj)
        self.delete(obj)
        return paths # the paths is the receipt in this implementation


    def withdraw(self, obj, receipt):
        """
        Used to de-relate an object after a move in a "manage_afterAdd()" hook.
        When given an object and a receipt, it gets the relations belonging to
        the receipt out of the deposit, and relates them to the object.
        self.relations.withDraw(self, receipt)
        """
        self.relateByPaths(obj, receipt)


    ########################################################
    # Methods used to display the relations


    def rootObjects(self, meta_types=None):
        roots = self.pathsToObjects(self.relations.keys(), meta_types)
        return roots


    def index_html(self):
        """
        Renders the relations in html. The result can get quite big 
        with many relations (several MBytes), so this is the most 
        efficient approach.
        If it should be inserted into a dtml or zpt document, it would 
        at least double the needed memory. I know this for a fact :-(
        """
        title = self.title_or_id()
        header = """
            <html>
            <head>
            <title>Relations for: %s</title>
            </head>
            <body>
            <h2>Relations for: %s</h2>
        """ % (title,title)
        footer = """
            </body>
            </html>
        """
        out = StringIO()
        out.write(header)
        out.write('<b>id:</b> %s<br>\n' % self.getId())
        out.write('<b>title:</b> %s<br><br>\n' % self.title_or_id())
        objs = self.rootObjects()
        for obj in objs:
            out.write('<b><a href="%s">%s</a></b><br>\n' %
                    (obj.absolute_url(), obj.title_or_id()))
            for subObj in self.get(obj):
                out.write('<a href="%s">%s</a><br>\n' %
                        (subObj.absolute_url(), subObj.title_or_id()))
        out.write(footer)
        return out.getvalue()


##########################
# constructor pages. Only used when the product is added to a folder.

def manage_addmxmRelationsAction(self, id='relations', title='Relations', REQUEST=None):
    "Add a mxmRelations to a folder."
    self._setObject(id, mxmRelations(id, title))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

# Get user input from this form
manage_addmxmRelationsForm = DTMLFile('manage_addmxmRelationsForm', globals())

